#include <u.h>
#include <libc.h>
#include <fcall.h>
#include <stdio.h>

void
error(char *fmt, ...)
{
	va_list arg;
	char buf[256];
	Fmt f;

	fmtfdinit(&f, 2, buf, sizeof buf);
	fmtprint(&f, "%s: ", argv0);
	va_start(arg, fmt);
	fmtprint(&f, fmt, arg);
	va_end(arg);
	fmtprint(&f, "\n");
	fmtfdflush(&f);
	exits(fmt);
}

void*
emalloc(uint n)
{
	void *p;

	p = malloc(n);
	if(p == nil)
		error("can't malloc: %r");
	memset(p, 0, n);
	return p;
}

char*
estrdup(char *s)
{
	char *t;

	t = emalloc(strlen(s)+1);
	strcpy(t, s);
	return t;
}

char*
estrstrdup(char *s, char *t)
{
	char *u;

	u = emalloc(strlen(s)+strlen(t)+1);
	strcpy(u, s);
	strcat(u, t);
	return u;
}

char*
estrstrstrdup(char *r, char *s, char *t)
{
	char *u;

	u = emalloc(strlen(r)+strlen(s)+strlen(t)+1);
	strcpy(u, r);
	strcat(u, s);
	strcat(u, t);
	return u;
}

int
gethi(char *dir)
{
	Dir *d;
	int hi;

	if((d = dirstat(dir)) == nil)
		return -1;
	hi = d->qid.vers;
	free(d);
	return hi;
}

void
usage(void)
{
	fprint(2, "usage: gethi [-d /mnt/news] comp.os.plan9\n");
	exits("usage");
}

void
main(int argc, char **argv)
{
	char *p, *q;
	Dir *d;
	char *dir = "/mnt/news";
	char *group;

	ARGBEGIN{
	case 'd':
		dir = EARGF(usage());
		break;
	default:
		usage();
		break;
	}ARGEND

	if(argc != 1)
		usage();

	group = estrdup(argv[0]);	/* someone will be cute */
	while(q=strchr(group, '/'))
		*q = '.';

	p = estrdup(argv[0]);
	while(q=strchr(p, '.'))
		*q = '/';
	p = estrstrstrdup(dir, "/", p);
	cleanname(p);

	if((d = dirstat(p)) == nil){	/* maybe it is a new group */
		if((d = dirstat(dir)) == nil){
			fprint(2, "dirstat(%s) fails: %r\n", dir);
			exits(0);
		}
		if((d->mode&DMDIR)==0){
			fprint(2, "%s not a directory\n", dir);
			exits(0);
		}
		free(d);
		if((d = dirstat(p)) == nil){
			fprint(2, "stat %s: %r\n", p);
			exits(0);
		}
	}
	if((d->mode&DMDIR)==0){
		fprint(2, "%s not a directory\n", dir);
		exits(0);
	}
	free(d);
	dir = estrstrdup(p, "/");

	print("%d\n", gethi(dir));
	exits(0);
}
